/*
dttools is Copyright (C) 2002 Douglas Thain
This software is released under the GNU General Public License.
See the file COPYING for details.
*/

#include "debug.h"
#include "domain_name_cache.h"
#include "datagram.h"
#include "macros.h"
#include "stringtools.h"
#include "full_io.h"
#include "xmalloc.h"
#include "full_io.h"

#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <sys/time.h>
#include <fcntl.h>
#include <errno.h>

static int debug_fd = 2;
static char *debug_file = 0;
static int debug_file_size = 10485760;
static char debug_addr[DATAGRAM_ADDRESS_MAX];
static struct datagram *debug_port=0;
static char *program_name = "";
static int debug_flags = D_NOTICE|D_RAW;
static int (*debug_getpid)() = getpid;

struct flag_info {
	const char *name;
	int flag;
};

static struct flag_info table[] = {
	{ "syscall",  D_SYSCALL },
	{ "notice",   D_NOTICE },
	{ "channel",  D_CHANNEL },
	{ "process",  D_PROCESS },
	{ "resolve",  D_RESOLVE },
	{ "libcall",  D_LIBCALL },
	{ "tcp",      D_TCP },
	{ "dns",      D_DNS },
	{ "auth",     D_AUTH },
	{ "local",    D_LOCAL },
	{ "http",     D_HTTP },
	{ "ftp",      D_FTP },
	{ "nest",     D_NEST },
	{ "kangaroo", D_KANGAROO },
	{ "chirp",    D_CHIRP },
	{ "dcap",     D_DCAP },
	{ "rfio",     D_RFIO },
	{ "cache",    D_CACHE },
	{ "poll",     D_POLL },
	{ "remote",   D_REMOTE },
	{ "debug",    D_DEBUG },
	{ "time",     D_TIME },
	{ "pid",      D_PID },
	{ "all",      ~0 },
	{ 0,0 }
};

struct fatal_callback {
	void (*callback) ();
	struct fatal_callback *next;
};

struct fatal_callback * fatal_callback_list = 0;


int  debug_flags_set( const char *flagname )
{
	struct flag_info *i;

	for(i=table;i->name;i++) {
		if(!strcmp(flagname,i->name)) {
			debug_flags |= i->flag;
			return 1;
		}
	}

	return 0;
}

void debug_flags_print( FILE *stream )
{
	int i;

	for(i=0;table[i].name;i++) {
		fprintf(stream,"%s ",table[i].name);
	}
}

static const char * flag_to_name( int flag )
{
	struct flag_info *i;

	for(i=table;i->name;i++) {
		if(i->flag&flag) return i->name;
	}

	return "debug";
}

static void do_debug( int is_fatal, int flags, char *fmt, va_list args )
{
	char newfmt[DATAGRAM_PAYLOAD_MAX];
	char buffer[DATAGRAM_PAYLOAD_MAX];
	int length;
	int pos =0;

	if(flags&D_RAW) {
		strcpy(newfmt,fmt);
	} else {
		if(debug_flags&D_TIME) {
			struct timeval tv;
			gettimeofday(&tv,0);
			pos += sprintf(newfmt,"%ld.%06ld ",tv.tv_sec,tv.tv_usec);
		}
		if(debug_flags&D_PID) {
			pos += sprintf(&newfmt[pos],"[%d] ",debug_getpid());
		}
		pos += sprintf(&newfmt[pos],"%s: %s: %s",program_name,is_fatal ? "fatal " : flag_to_name(flags),fmt);
	}

	vsprintf(buffer,newfmt,args);
	string_chomp(buffer);
	strcat(buffer,"\n");
	length = strlen(buffer);

	if(debug_port) {
		datagram_send(debug_port,buffer,length,debug_addr,DEBUG_PORT);
	} else {
		full_write(debug_fd,buffer,length);
		if(debug_file) {
			int size = lseek(debug_fd,0,SEEK_END);
			if( size >= debug_file_size && debug_file_size!=0 ) {
				char *newname = alloca(strlen(debug_file)+5);
				sprintf(newname,"%s.old",debug_file);
				rename(debug_file,newname);
				full_write(debug_fd,"*** rotating file ***\n",22);
				close(debug_fd);
				debug_fd = open(debug_file,O_CREAT|O_TRUNC|O_WRONLY,0777);
				if(debug_fd<0) fatal("couldn't open %s: %s",debug_file,strerror(errno));
			}
		}
	}
}

void debug( int flags, char *fmt, ... )
{
	va_list args;
	va_start(args,fmt);

	if(flags&debug_flags) {
		do_debug(0,flags,fmt,args);
	}

	va_end(args);
}

void fatal( char *fmt, ... ) 
{
	struct fatal_callback *f;
	va_list args;
	va_start(args,fmt);

	do_debug(1,0,fmt,args);

	for(f=fatal_callback_list;f;f=f->next) {
		f->callback();
	}

	while(1) kill(getpid(),SIGKILL);

	va_end(args);
}

void debug_config_fatal( void (*callback) () )
{
	struct fatal_callback *f;
	f = xmalloc(sizeof(*f));
	f->callback = callback;
	f->next = fatal_callback_list;
	fatal_callback_list = f;
}

void debug_config( char *name )
{
	char *host = getenv("DEBUG_HOST");
	char *file = getenv("DEBUG_FILE");
	char *filesize = getenv("DEBUG_FILE_SIZE");

	program_name = strrchr(name,'/');
	if(program_name) {
		program_name++;
	} else {
		program_name = name;
	}

	if(host) debug_config_host(host);
	if(file) debug_config_file(file);
	if(filesize) debug_config_file_size(atoi(filesize));
}

void debug_config_file( const char *f )
{
	if(debug_file) {
		free(debug_file);
		debug_file = 0;
	}

	if(f) {
		debug_file = strdup(f);
		debug_fd = open(f,O_CREAT|O_APPEND|O_WRONLY,0777);
		if(debug_fd<0) fatal("couldn't open %s: %s",f,strerror(errno));		
	} else {
		debug_fd = 2;
	}
}

void debug_config_file_size( int size )
{
	debug_file_size = size;
}

int debug_config_host( const char *host )
{
	char addr[DATAGRAM_ADDRESS_MAX];

	if(host) {
		if(!domain_name_cache_lookup(host,addr)) return 0;
		strcpy(debug_addr,addr);
		if(!debug_port) debug_port = datagram_create(DATAGRAM_PORT_ANY);
		debug_flags = debug_flags|D_TIME|D_PID;
		return debug_port!=0;

	} else {
		if(debug_port) {
			datagram_delete(debug_port);
			debug_port = 0;
		}
		return 1;
	}
}

void debug_config_getpid( int (*getpidfunc)() )
{
	debug_getpid = getpidfunc;
}

int debug_flags_clear()
{
	int result = debug_flags;
	debug_flags = 0;
	return result;	
	debug_flags = 0;
}

void debug_flags_restore( int fl )
{
	debug_flags = fl;
}

